<!doctype html>
<meta http-equiv="Content-Type" content="text/html; charset={{ encoding }}">

{# Strip out the default script files we provide, but keep some, like mathjax. #}
{% set script_files = [] %}
{%- set url_root = pathto('', 1) %}
{%- if url_root == '#' %}{% set url_root = '' %}{% endif %}
{%- set titlesuffix = " &mdash; "|safe + docstitle|e %}

{%- macro sidebar() %}
<div class="sphinxsidebar">
<div class="sphinxsidebarwrapper" style="padding-top: 1em;">
    {%- if logo %}
      <p class="logo"><a href="{{ pathto(master_doc) }}">
        <img class="logo" src="{{ pathto('_static/' + logo, 1) }}" alt="">
      </a></p>
    {%- endif %}

    {% if theme_github_repo %}
    <p>
    <a href="{{ theme_github_repo }}">GitHub Repository</a>
    </p>
    {% endif %}

    <p>
    <a href="/en/release/">Docs for last release</a>
    </p>

    <p>
    <a href="/en/latest/">Docs for master branch</a>
    </p>

    {%- block sidebartoc %}
    {%- include "localtoc.html" %}
    {%- endblock %}
    {%- block sidebarrel %}
    {%- include "relations.html" %}
    {%- endblock %}
</div>
</div>
{%- endmacro %}

<title>{{ title|striptags|e }}{{ titlesuffix }}</title>
<link rel="stylesheet" href="{{ pathto('_static/' + style, 1) }}">
<link rel="stylesheet" href="{{ pathto('_static/pygments.css', 1) }}">
{%- for cssfile in css_files %}
    <link rel="stylesheet" href="{{ pathto(cssfile, 1) }}">
{%- endfor %}

<script>
var DOCUMENTATION_OPTIONS = {
    URL_ROOT:    '{{ url_root }}',
    VERSION:     '{{ release|e }}',
    COLLAPSE_INDEX: false,
    FILE_SUFFIX: '{{ '' if no_search_suffix else file_suffix }}',
    HAS_SOURCE:  {{ has_source|lower }}
};
var doc_version = "{{ current_version }}";
var doc_slug = "{{ slug }}";
var page_name = "{{ pagename }}";
</script>

{%- for scriptfile in script_files %}
<script src="{{ pathto(scriptfile, 1) }}"></script>
{%- endfor %}


{%- block content %}
  {%- block sidebar1 %} {# possible location for sidebar #} {% endblock %}

    <div class="document">
  {%- block document %}
      <div class="documentwrapper">
        <div class="bodywrapper">
          <div class="body">
            {% block body %} {% endblock %}
          </div>
        </div>
      </div>
  {%- endblock %}

  {%- block sidebar2 %}{{ sidebar() }}{% endblock %}
      <div class="clearer"></div>
    </div>
{%- endblock %}

{% macro relbar() %}{% endmacro %}
{%- block relbar2 %}{{ relbar() }}{% endblock %}


<div class="footer">
{%- if show_copyright %}
    {%- if hasdoc('copyright') %}
    {% trans path=pathto('copyright'), copyright=copyright|e %}&copy; <a href="{{ path }}">Copyright</a> {{ copyright }}.{% endtrans %}
  {%- else %}
    {% trans copyright=copyright|e %}&copy; Copyright {{ copyright }}.{% endtrans %}
  {%- endif %}
{%- endif %}
{%- if last_updated %}
  {% trans last_updated=last_updated|e %}Last updated on {{ last_updated }}.{% endtrans %}
{%- endif %}
{%- if show_sphinx %}
  {% trans sphinx_version=sphinx_version|e %}Created using <a href="http://sphinx.pocoo.org/">Sphinx</a> {{ sphinx_version }}.{% endtrans %}
{%- endif %}
</div>

{% if not using_theme %}
    {# Keep this here, so that the RTD logo doesn't stomp on the bottom of the theme. #}
    <br/>
    <br/>
    <br/>
{% endif %}


<script>
  var _paq = _paq || [];
  _paq.push(['trackPageView']);
  _paq.push(['enableLinkTracking']);
  (function() {
    var u=(("https:" == document.location.protocol) ? "https" : "http") + "://37.187.23.96/tr/";
    _paq.push(['setTrackerUrl', u+'ep']);
    _paq.push(['setSiteId', 2]);
    var d=document, g=d.createElement('script'), s=d.getElementsByTagName('script')[0]; g.type='text/javascript';
    g.defer=true; g.async=true; g.src=u+'js'; s.parentNode.insertBefore(g,s);
  })();
</script>
<noscript><p><img src="http://37.187.23.96/tr/ep?idsite=2" style="border:0" alt=""></p></noscript>
